#!/usr/bin/env python
# -*- encoding: utf-8 -*-

import xml.etree.ElementTree as ET

from .binary import BinaryModel

class PseudoDolloModel(BinaryModel):
    package_notice = """[DEPENDENCY]: The Pseudo-Dollo model is implemented in the BEAST package "Babel"."""

    def __init__(self, model_config, global_config):

        BinaryModel.__init__(self, model_config, global_config)
        self.datatype_id = None
        self.subst_model_id = None
#        self.freq_str = self.build_freq_str()
        
    def JUNK_build_freq_str(self):
        if self.binarised:
            all_data = []
            for f in self.features:
                for lang in self.data:
                    if self.data[lang][f] == "?":
                        continue
                    dpoint, index = self.data[lang][f], self.unique_values[f].index(self.data[lang][f])
                    all_data.append(index)
        else:
            all_data = []
            for f in self.features:
                for lang in self.data:
                    if self.data[lang].get(f,"?") == "?":
                        valuestring = "".join(["?" for i in range(0,len(self.unique_values[f])+1)])
                    else:
                        valuestring = ["0" for i in range(0,len(self.unique_values[f])+1)]
                        valuestring[self.unique_values[f].index(self.data[lang][f])+1] = "1"
                        all_data.extend(valuestring)

        all_data = [d for d in all_data if d !="?"]
        all_data = [int(d) for d in all_data]
        zerf = 1.0*all_data.count(0) / len(all_data)
        onef = 1.0*all_data.count(1) / len(all_data)
        assert abs(1.0 - (zerf+onef)) < 1e-6
        return "%.2f %.2f" % (zerf, onef)

    def add_state(self, state):
        BinaryModel.add_state(self, state)
        ET.SubElement(state, "parameter", {"id":"PDCognateDeathRate", "name":"stateNode", "upper":"1.0"}).text = "0.05"
        ET.SubElement(state, "parameter", {"id":"PDFreqParameter", "dimension":"3", "lower":"0.0", "name":"stateNode", "upper":"1.0"}).text = "0.95 0.03 0.02 0.00"

    def get_userdatatype(self, feature, fname):
        if self.datatype_id:
            return ET.Element("userDataType", {"idref":self.datatype_id})
        else:
            self.datatype_id = "bdDataType"
            datatype = ET.Element("userDataType", {
                "id":self.datatype_id,
                "spec":"beast.evolution.datatype.UserDataType",
                "states":"4",
                "codelength":"1",
                "codeMap": "A = 0, 1 = 1, B = 2, 0 = 0 2, ? = 0 1 2, - = 0 1 2, C = 0 1 2, D = 3",
                })
            return datatype

    def add_sitemodel(self, distribution, feature, fname):

        # Sitemodel
        mr = self.get_mutation_rate(feature, fname)
        sitemodel = ET.SubElement(distribution, "siteModel", {"id":"SiteModel.%s"%fname,"spec":"SiteModel", "mutationRate":mr,"shape":"1","proportionInvariant":"0"})
        substmodel = self.add_substmodel(sitemodel, feature, fname)

    def add_substmodel(self, sitemodel, feature, fname):
        if self.subst_model_id:
            sitemodel.set("substModel", "@%s" % self.subst_model_id)
            return

        self.subst_model_id = "%s:pseudodollo.s" % self.name
        substmodel = ET.SubElement(sitemodel, "substModel",{"id":self.subst_model_id,"spec":"beast.evolution.substitutionmodel.BirthDeathModel", "deathprob":"@PDCognateDeathRate"})
        ET.SubElement(substmodel, "frequencies", {"id":"PDfrequencies","spec":"Frequencies","frequencies":"@PDFreqParameter"})

    def add_prior(self, prior):
        BinaryModel.add_prior(self, prior)
        deathrate_prior = ET.SubElement(prior, "prior", {"id":"PDCognateDeathRatePrior", "name":"distribution","x":"@PDCognateDeathRate"})
        ET.SubElement(deathrate_prior, "Exponential", {"name":"distr","id":"%s:pd_deathrate_prior.s"%self.name, "mean":"1.0"} )

        freq_prior = ET.SubElement(prior, "prior", {"id":"PDFreqParameterPrior", "name":"distribution", "x":"@PDFreqParameter"})
        dirich = ET.SubElement(freq_prior, "distr", {"id":"PDFreqUniform", "spec":"beast.math.distributions.Uniform", "lower":"0.0", "upper":"1.0"})

    def add_operators(self, run):
        BinaryModel.add_operators(self, run)
        ET.SubElement(run, "operator", {"id":"pd_deathrate_scaler.s", "spec":"ScaleOperator","parameter":"@PDCognateDeathRate","scaleFactor":"0.75","weight":"3.0"})
        ET.SubElement(run, "operator", {"id":"pd_frequencies_exchanger", "spec":"DeltaExchangeOperator","parameter":"@PDFreqParameter", "weight":"3.0"})

    def add_param_logs(self, logger):
        BinaryModel.add_param_logs(self, logger)
#        ET.SubElement(logger,"log",{"idref":"%s:dollo_alpha.s" % self.name})
#        ET.SubElement(logger,"log",{"idref":"%s:dollo_s.s" % self.name})
#        if self.config.log_fine_probs:
#            ET.SubElement(logger,"log",{"idref":"%s:dollo_alpha_prior.s" % self.name})
#            ET.SubElement(logger,"log",{"idref":"%s:dollo_s_prior.s" % self.name})



#!/usr/bin/env python

try:
    from setuptools import setup
except ImportError:
    from distutils.core import setup

from beastling import __version__ as version


requires = [
    'six',
    'newick>=0.6.0',
    'appdirs',
    'clldutils~=2.0',
    'pycldf',
]

setup(
    name='beastling',
    version=version,
    description='Command line tool to help mortal linguists use BEAST',
    author='Luke Maurits',
    author_email='luke@maurits.id.au',
    license="BSD (3 clause)",
    classifiers=[
        'Programming Language :: Python',
        "Programming Language :: Python :: 2",
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'License :: OSI Approved :: BSD License',
    ],
    packages=['beastling','beastling.clocks','beastling.fileio','beastling.models'],
    install_requires=requires,
    tests_require=['mock==1.0.0', 'nose'],
    entry_points={
        'console_scripts': ['beastling=beastling.cli:main'],
    },
    package_data={'beastling': ['data/*']},
)



import io
import os
import collections
import xml.etree.ElementTree as ET

from ..fileio.datareaders import load_data

class BaseModel(object):
    """
    Base class from which all substitution model classes are descended.
    Implements generic functionality which is common to all substitution
    models, such as rate variation.
    """

    treewide_reconstruction = False
    """Should ASR be performed on the entire tree (if at all)?"""

    def __init__(self, model_config, global_config):
        """
        Parse configuration options, load data from file and pre-process data.
        """
        self.messages = []
        self.config = global_config

        self.name = model_config["name"] 
        self.data_filename = model_config["data"] 
        self.clock = model_config.get("clock", "")
        self.features = model_config.get("features",["*"])
        self.reconstruct = model_config.get("reconstruct", None)
        self.reconstruct_at = model_config.get("reconstruct_at", [])
        self.exclusions = model_config.get("exclusions",None)
        self.constant_feature = False
        self.constant_feature_removed = False
        self.frequencies = model_config.get("frequencies", "empirical")
        self.pruned = model_config.get("pruned", False)
        self.rate_variation = model_config.get("rate_variation", False)
        self.feature_rates = model_config.get("feature_rates", None)
        self.rate_partition = model_config.get("rate_partition", None)
        self.ascertained = model_config.get("ascertained", None)
        # Force removal of constant features here
        # This can be set by the user in BinaryModel only
        self.remove_constant_features = True
        self.minimum_data = float(model_config.get("minimum_data", 0))
        self.substitution_name = self.__class__.__name__
        self.data_separator = ","
        self.use_robust_eigensystem = model_config.get("use_robust_eigensystem", False)
        self.metadata = []
        self.treedata = []

        # Load the entire dataset from the file
        self.data = load_data(self.data_filename, file_format=model_config.get("file_format",None), lang_column=model_config.get("language_column",None), value_column=model_config.get("value_column",None))
        # Remove features not wanted in this analysis
        self.build_feature_filter()
        self.apply_feature_filter()

        # Keep this around for later...
        self.global_config = global_config

    def build_feature_filter(self):
        """
        Create the self.feature_filter attribute, which is a set of feature
        names that functions analogously to Configuration.lang_filter
        attribute.
        """
        if self.features == ["*"]:
            random_iso = list(self.data.keys())[0]
            self.features = set()
            for lang_features in self.data.values():
                self.features |= set(lang_features.keys())
            self.features = list(self.features)
        if self.exclusions:
            self.features = [f for f in self.features if f not in self.exclusions]
        self.feature_filter = set(self.features)

        if self.reconstruct == ["*"]:
            self.reconstruct = self.features[:]
        elif self.reconstruct:
            fail_to_find = [f for f in self.reconstruct if f not in self.features]
            if fail_to_find:
                self.messages.append(
                    "[WARNING] Model {:s}:"
                    " Features {:} not found, cannot be reconstructed.""".format(
                        self.name, fail_to_find))
            self.reconstruct = [f for f in self.reconstruct if f in self.features]
            self.messages.append(
                    "[INFO] Model \"{:s}\":"
                    " Features {:} will be reconstructed.""".format(
                        self.name, self.reconstruct))
            # Note: That is a lie. Features can still be filtered out by
            # subsequent decisions, eg. because they are constant.
        else:
            self.reconstruct = []

        if self.reconstruct_at == ["*"]:
            self.reconstruct_at = None
            self.treewide_reconstruction = True
        elif self.reconstruct_at:
            if len(self.reconstruct_at) > 1:
                raise ValueError("Cannot currently reconstruct at more than one location.")
            for f in self.reconstruct_at:
                if f not in self.config.language_group_configs:
                    raise KeyError("Language group {:} is undefined. Valid groups are: {:}".format(
                        f, ", ".join(self.config.language_groups.keys())))
        elif self.reconstruct:
            self.reconstruct_at=["root"]

    def process(self):
        """
        Subsample the data set to include only those languages and features
        which are compatible with the settings.
        """
        self.apply_language_filter()
        self.compute_feature_properties()
        self.remove_unwanted_features()
        self.load_rate_partition()
        if self.rate_partition:
            self.all_rates = sorted(list(set(self.rate_partition.values())))
        elif self.rate_variation or self.feature_rates:
            self.all_rates = self.features
        self.load_feature_rates()
        if self.rate_partition and not (self.feature_rates or self.rate_variation):
            self.messages.append("""[WARNING] Model %s: Estimating rates for feature partitions because no fixed rates were provided, is this what you wnated?  Use rate_variation=True to make this implicit.""" % self.name)
            self.rate_variation = True
        self.compute_weights()
        if self.pruned:
            self.messages.append("""[DEPENDENCY] Model %s: Pruned trees are implemented in the BEAST package "BEASTlabs".""" % self.name)

    def apply_language_filter(self):
        """
        Remove all languages from the data set which are not part of the
        configured language filter.
        """
        all_langs = self.data.keys()
        langs_to_remove = [l for l in all_langs if not self.config.filter_language(l)]
        for lang in langs_to_remove:
            self.data.pop(lang)
        # Make sure we've not removed all languages
        if not self.data.keys():
            raise ValueError("Language filters leave nothing in the dataset for model '%s'!" % self.name)
        # Keep a sorted list so that the order of things in XML is deterministic
        self.languages = sorted(list(self.data.keys()))

    def load_rate_partition(self):
        """
        Load a partition of features for sharing mutation rates.
        """
        if not self.rate_partition:
            self.rate_partition = {}
            return
        if not os.path.exists(self.rate_partition):
            raise ValueError("Could not find feature rate file %s." % self.rate_partition)
        fname = self.rate_partition
        with open(fname) as fp:
            self.rate_partition = {}
            for line in fp:
                name, part = line.split(":",1)
                name = name.strip()
                part = [p.strip() for p in part.split(",")]
                part = [p for p in part if p in self.features]
                for p in part:
                    self.rate_partition[p] = name

    def load_feature_rates(self):
        """
        Load relative feature rates from .csv file.
        """
        if not self.feature_rates:
            self.feature_rates = {}
            return
        if not os.path.exists(self.feature_rates):
            raise ValueError("Could not find feature rate file %s." % self.feature_rates)
        fname = self.feature_rates
        with io.open(self.feature_rates, encoding="UTF-8") as fp:
            self.feature_rates = {}
            for line in fp:
                feature, rate = line.split(",")
                feature = feature.strip()
                # Skip irrelevant things
                if feature not in self.all_rates:
                    continue
                rate = float(rate.strip())
                self.feature_rates[feature] = rate
        if not all((rate in self.feature_rates for rate in self.all_rates)):
            self.messages.append("""[WARNING] Model "%s": Rate file %s does not contain rates for every feature/partition.  Missing rates will default to 1.0, please check that this is okay.""" % (self.name, fname))
        if not self.feature_rates:
            self.messages.append("""[WARNING] Model "%s": Could not find any valid feature or partition rates in the file %s, is this the correct file for this analysis?""" % (self.name, fname))
            return
        norm = sum(self.feature_rates.values()) / len(self.feature_rates.values())
        for f in self.feature_rates:
            self.feature_rates[f] /= norm

    def apply_feature_filter(self):
        """
        Remove all features from the data set which are not part of the
        configured feature filter.
        """
        self.features = set()
        for language in self.data.values():
            features_in_data = set(language.keys())
            features_to_keep = features_in_data & self.feature_filter
            self.features |= features_to_keep
            features_to_remove = features_in_data - features_to_keep
            for feat in features_to_remove:
                language.pop(feat)
        self.features = sorted(list(self.features))

    def compute_feature_properties(self):
        """
        Compute various items of metadata for all remaining features.
        """

        self.valuecounts = {}
        self.extracolumns = collections.defaultdict(int)
        self.unique_values = {}
        self.missing_ratios = {}
        self.counts = {}
        self.codemaps = {}
        for f in self.features:
            # Compute various things
            all_values = [self.data[l].get(f,"?") for l in self.data]
            missing_data_ratio = all_values.count("?") / (1.0*len(all_values))
            non_q_values = [v for v in all_values if v != "?"]
            counts = {}
            for v in non_q_values:
                counts[v] = non_q_values.count(v)
            unique_values = list(set(non_q_values))
            # Sort unique_values carefully.
            # Possibly all feature values are numeric strings, e.g. "1", "2", "3".
            # If we sort these as strings then we get weird things like "10" < "2".
            # This can actually matter for things like ordinal models.
            # So convert these to ints first...
            if all([v.isdigit() for v in unique_values]):
                unique_values = list(map(int, unique_values))
                unique_values.sort()
                unique_values = list(map(str, unique_values))
            # ...otherwise, just sort normally
            else:
                unique_values.sort()
            self.unique_values[f] = unique_values

            N = len(unique_values)
            self.valuecounts[f] = N
            self.missing_ratios[f] = missing_data_ratio
            self.counts[f] = counts
            self.codemaps[f] = self.build_codemap(unique_values)

    def remove_unwanted_features(self):
        """
        Remove any undesirable features from the dataset, such as those with
        no data for the configured set of languages, constant features, etc.
        """

        bad_feats = []
        for f in self.features:

            # Exclude features with no data
            if self.valuecounts[f] == 0:
                self.messages.append("""[INFO] Model "%s": Feature %s excluded because there are no datapoints for selected languages.""" % (self.name, f))
                bad_feats.append(f)
                continue

            # Exclude features with lots of missing data
            missing_ratio = self.missing_ratios[f]
            if int(100*(1.0-missing_ratio)) < self.minimum_data:
                self.messages.append("""[INFO] Model "%s": Feature %s excluded because of excessive missing data (%d%%).""" % (self.name, f, int(missing_ratio*100)))
                bad_feats.append(f)
                continue

            # Exclude constant features
            if self.valuecounts[f] == 1:
                if self.remove_constant_features:
                    self.constant_feature_removed = True
                    self.messages.append("""[INFO] Model "%s": Feature %s excluded because its value is constant across selected languages.  Set "remove_constant_features=False" in config to stop this.""" % (self.name, f))
                    bad_feats.append(f)
                    continue
                else:
                    self.constant_feature = True

        for bad in bad_feats:
            self.features.remove(bad)
            for lang in self.languages:
                if bad in self.data[lang]:
                    self.data[lang].pop(bad)

        # Make sure there's something left
        if not self.features:
            raise ValueError("No features specified for model %s!" % self.name)
        self.features.sort()
        self.messages.append("""[INFO] Model "%s": Using %d features from data source %s""" % (self.name, len(self.features), self.data_filename))
        if self.constant_feature and self.rate_variation:
            self.messages.append("""[WARNING] Model "%s": Rate variation enabled with constant features retained in data.  This *may* skew rate estimates for non-constant features.""" % self.name)

    def compute_weights(self):
        self.weights = []
        # Weights feed into a DeltaExchangeOperator, so they need to
        # be integers. This is currently implicit, not enforced.
        if self.rate_partition:
            parts = list(self.rate_partition.values())
            partition_weights = {p:parts.count(p) for p in parts}
            for part in sorted(list(set(self.rate_partition.values()))):
                self.weights.append(partition_weights[part])
        else:
            for f in self.features:
                self.weights.append(1)

    def set_ascertained(self):
        """
        Decide whether or not to do ascertainment correction for non-constant
        data, unless the user has provided an explicit setting.
        """

        # Do the correction if the tree is calibrated, as ascertainment
        # correction influences timing estimates
        if self.ascertained == None:
            self.ascertained = not self.global_config.arbitrary_tree

    def build_codemap(self, unique_values):
        """
        Build a codemap string for a feature.
        """
        N = len(unique_values)
        codemapbits = []
        codemapbits.append(",".join(["%d=%d" % (n,n) for n in range(0,len(unique_values))]))
        codemapbits.append("?=" + " ".join([str(n) for n in range(0,N)]))
        codemapbits.append("-=" + " ".join([str(n) for n in range(0,N)]))
        return ",".join(codemapbits)

    def add_misc(self, beast):
        pass

    def add_state(self, state):
        """Construct the model's state nodes.

        Add parameters for Gamma-distributed rate heterogenetiy, if
        configured.

        """

        if self.frequencies == "estimate":
            self.add_frequency_state(state)

        if self.rate_variation:
            if self.feature_rates:
                # If user specified rates have been provided for either
                # features or partitions, we need to list each rate
                # individually
                for rate in self.all_rates:
                    param = ET.SubElement(state, "parameter", {
                        "id":"featureClockRate:%s:%s" % (self.name, rate),
                        "name":"stateNode"})
                    param.text=str(self.feature_rates.get(rate,1.0))
            else:
                # If not, and everything is initialised to the same
                # value, we can just whack 'em all in a big plate
                plate = ET.SubElement(state, "plate", {
                    "var":"rate",
                    "range":",".join(self.all_rates)})
                param = ET.SubElement(plate, "parameter", {
                    "id":"featureClockRate:%s:$(rate)" % self.name,
                    "name":"stateNode"})
                param.text="1.0"

            # Give Gamma shape parameter a finite domain
            # Must be > 1.0 for the distribution to be bell-shaped,
            # rather than L-shaped.  The domain [1.1,1000] limits feature
            # rate variation to the realms of vague plausibity
            parameter = ET.SubElement(state, "parameter", {"id":"featureClockRateGammaShape:%s" % self.name, "lower":"1.1","upper":"100.0","name":"stateNode"})
            parameter.text="5.0"
            # Gamma scale parameter's domain is defined *implicilty*
            # by the fact that the operators maintain shape*scale = 1.0
            parameter = ET.SubElement(state, "parameter", {"id":"featureClockRateGammaScale:%s" % self.name, "name":"stateNode"})
            parameter.text="0.2"

    def add_frequency_state(self, state):
        for f in self.features:
            fname = "%s:%s" % (self.name, f)
            param = ET.SubElement(state,"stateNode",{
                "id":"feature_freqs_param.s:%s"%fname,
                "spec":"parameter.RealParameter",
                "dimension":str(self.valuecounts[f]),
                "lower":"0.0",
                "upper":"1.0",
            })
            param.text = str(1.0/self.valuecounts[f])

    def add_prior(self, prior):
        """
        Add prior distributions for Gamma-distributed rate heterogenetiy, if
        configured.
        """
        if self.rate_variation:
            # Gamma prior with mean 1 over all mutation rates
            sub_prior = ET.SubElement(prior, "prior", {"id":"featureClockRatePrior.s:%s" % self.name, "name":"distribution"})
            compound = ET.SubElement(sub_prior, "input", {"id":"featureClockRateCompound:%s" % self.name, "spec":"beast.core.parameter.CompoundValuable", "name":"x"})
            plate = ET.SubElement(compound, "plate", {
                "var":"rate",
                "range":",".join(self.all_rates)})
            ET.SubElement(plate, "var", {
                "idref":"featureClockRate:%s:$(rate)" % self.name})
            gamma  = ET.SubElement(sub_prior, "input", {"id":"featureClockRatePriorGamma:%s" % self.name, "spec":"beast.math.distributions.Gamma", "name":"distr", "alpha":"@featureClockRateGammaShape:%s" % self.name, "beta":"@featureClockRateGammaScale:%s" % self.name})
            # Exponential hyperprior on scale of Gamma prior
            # Exponential prior favours small scales over large scales, i.e. less rate variation
            # Mean scale 0.23 chosen for general sensibility, e.g.:
            #   - Prior distribution is roughly 50/50 that ratio of fastest
            #     to slowest feature rate in a dataset of size 200 is below
            #     or above 10.
            #   - Prior probability of roughly 0.90 that this ratio is below
            #     100.
            sub_prior = ET.SubElement(prior, "prior", {"id":"featureClockRateGammaScalePrior.s:%s" % self.name, "name":"distribution", "x":"@featureClockRateGammaScale:%s" % self.name})
            ET.SubElement(sub_prior, "Exponential", {"id":"featureClockRateGammaShapePriorExponential.s:%s" % self.name, "mean":"0.23", "name":"distr"})

    def pattern_names(self, feature):
        """Content of the columns corresponding to this feature in the alignment.

        This method is used for displaying helpful column names in ancestral
        state reconstruction output. It gives column headers for actual value
        columns as well as for dummy columns used in ascertainment correction,
        if such columns exist.

        """
        if self.ascertained:
            return ["{:}_dummy{:d}".format(feature, i)
                    for i in range(self.extracolumns[feature])] + [feature]
        else:
            return [feature]

    def add_likelihood(self, likelihood):
        """
        Add likelihood distribution corresponding to all features in the
        dataset.
        """
        for n, f in enumerate(self.features):
            fname = "%s:%s" % (self.name, f)
            attribs = {"id": "featureLikelihood:%s" % fname,
                       "spec": "TreeLikelihood",
                       "useAmbiguities": "true"}

            if self.pruned:
                distribution = ET.SubElement(likelihood, "distribution",attribs)
                # Create pruned tree
                tree_id = "Tree.t:prunedBeastlingTree.%s" % fname
                tree = ET.SubElement(distribution, "tree", {"id":tree_id, "spec":"beast.evolution.tree.PrunedTree","quickshortcut":"true","assert":"false"})
                ET.SubElement(tree, "tree", {"idref":"Tree.t:beastlingTree"})
                ET.SubElement(tree, "alignment", {"idref":"pruned_data_%s"%fname})
                # Create pruned branchrate
                self.clock.add_pruned_branchrate_model(distribution, fname, tree_id)
            else:
                attribs["branchRateModel"] = "@%s" % self.clock.branchrate_model_id
                attribs["tree"] = "@Tree.t:beastlingTree"
                distribution = ET.SubElement(likelihood, "distribution",attribs)

            if f in  self.reconstruct:
                # Use a different likelihood spec (also depending on whether
                # the whole tree is reconstructed, or only some nodes)
                if self.treewide_reconstruction:
                    distribution.attrib["spec"] = "AncestralStateTreeLikelihood"
                    self.treedata.append(attribs["id"])
                    distribution.attrib["tag"] = f
                else:
                    distribution.attrib["spec"] = "AncestralStateLogger"
                    distribution.attrib["value"] = " ".join(self.pattern_names(f))
                    for label in self.reconstruct_at:
                        langs = self.config.language_group(label)
                        self.beastxml.add_taxon_set(distribution, label, langs)
                    self.metadata.append(attribs["id"])
                distribution.attrib["useAmbiguities"] = "false"

            # Sitemodel
            self.add_sitemodel(distribution, f, fname)

            # Data
            self.add_feature_data(distribution, n, f, fname)

    def add_sitemodel(self, beast):
        pass

    def add_master_data(self, beast):
        self.filters = {}
        data = ET.SubElement(beast, "data", {
            "id":"data_%s" % self.name,
            "name":"data_%s" % self.name,
            "dataType":"integer"})
        for lang in self.languages:
            formatted_points = [self.format_datapoint(f, self.data[lang][f]) for f in self.features]
            value_string = self.data_separator.join(formatted_points)
            if not self.filters:
                n = 1
                for f, x in zip(self.features, formatted_points):
                    # Format the FilteredAlignment filter appropriately
                    if self.data_separator:
                        length = len(x.split(self.data_separator))
                    else:
                        length = len(x)
                    if length == 1:
                        self.filters[f] = str(n)
                    else:
                        self.filters[f] = "%d-%d" % (n, n+length-1)
                    n += length
            seq = ET.SubElement(data, "sequence", {
                "id":"data_%s:%s" % (self.name, lang),
                "taxon":lang,
                "value":value_string})

    def format_datapoint(self, feature, point):
        if self.ascertained:
            return self._ascertained_format_datapoint(feature, point)
        else:
            return self._standard_format_datapoint(feature, point)

    def _standard_format_datapoint(self, feature, point):
        if point == "?":
            return point
        else:
            return str(self.unique_values[feature].index(point))

    def _ascertained_format_datapoint(self, feature, point):
        extra_cols = self.valuecounts[feature]
        self.extracolumns[feature] = extra_cols
        if point == "?":
            return self.data_separator.join(["?" for i in range(0, extra_cols + 1)])
        else:
            cols = list(range(0, extra_cols))
            cols.append(self.unique_values[feature].index(point))
            return self.data_separator.join(map(str, cols))

    def add_feature_data(self, distribution, index, feature, fname):
        """
        Add <data> element corresponding to the indicated feature, descending
        from the indicated likelihood distribution.
        """
        if self.pruned:
            pruned_align = ET.SubElement(distribution,"data",{"id":"pruned_data_%s" % fname, "spec":"PrunedAlignment"})
            parent = pruned_align
            name = "source"
        else:
            parent = distribution
            name = "data"
        data = ET.SubElement(parent, name, {
            "id":"data_%s" % fname,
            "spec":"FilteredAlignment",
            "data":"@data_%s" % self.name,
            "filter":self.filters[feature]})
        if self.ascertained:
            data.set("ascertained", "true")
            data.set("excludefrom", "0")
            data.set("excludeto", str(self.valuecounts[feature]))
        data.append(self.get_userdatatype(feature, fname))
        return data

    def get_userdatatype(self, feature, fname):
        return ET.Element("userDataType", {"id":"featureDataType.%s"%fname,"spec":"beast.evolution.datatype.UserDataType","codeMap":self.codemaps[feature],"codelength":"-1","states":str(self.valuecounts[feature])})

    def get_mutation_rate(self, feature, fname):
        """
        Get a string which can be used as the mutationRate for a sitemodel.
        """
        if self.rate_variation:
            if self.rate_partition:
                mr = "@featureClockRate:%s:%s" % (self.name, self.rate_partition[feature])
            else:
                mr = "@featureClockRate:%s" % fname
        elif self.feature_rates:
            if self.rate_partition:
                mr = str(self.feature_rates.get(self.rate_partition[feature], 1.0))
            else:
                mr = str(self.feature_rates.get(feature, 1.0))
        else:
            mr = "1.0"
        return mr

    def add_operators(self, run):
        """
        Add <operators> for individual feature substitution rates if rate
        variation is configured.
        """
        if self.frequencies == "estimate":
            self.add_frequency_operators(run)
        if self.rate_variation:
            # UpDownOperator to scale the Gamma distribution for this model's
            # feature rates
            updown = ET.SubElement(run, "operator", {"id":"featureClockRateGammaUpDown:%s" % self.name, "spec":"UpDownOperator", "scaleFactor":"0.5","weight":"0.3"})
            ET.SubElement(updown, "parameter", {"idref":"featureClockRateGammaShape:%s" % self.name, "name":"up"})
            ET.SubElement(updown, "parameter", {"idref":"featureClockRateGammaScale:%s" % self.name, "name":"down"})

    def add_frequency_operators(self, run):
        for f in self.features:
            fname = "%s:%s" % (self.name, f)
            ET.SubElement(run, "operator", {"id":"estimatedFrequencyOperator:%s" % fname, "spec":"DeltaExchangeOperator", "parameter":"@feature_freqs_param.s:%s" % fname, "delta":"0.01","weight":"0.1"})

    def add_param_logs(self, logger):
        """
        Add entires to the logfile corresponding to individual feature
        substition rates if rate variation is configured.
        """
        if self.config.log_fine_probs:
            plate = ET.SubElement(logger, "plate", {
                "var":"feature",
                "range":",".join(self.features)})
            ET.SubElement(plate, "log", {
                "idref":"featureLikelihood:%s:$(feature)" % self.name})
            if self.rate_variation:
                ET.SubElement(logger,"log",{"idref":"featureClockRatePrior.s:%s" % self.name})
                ET.SubElement(logger,"log",{"idref":"featureClockRateGammaScalePrior.s:%s" % self.name})

        if self.frequencies == "estimate":
            self.add_frequency_logs(logger)
        if self.rate_variation:
            plate = ET.SubElement(logger, "plate", {
                "var":"rate",
                "range":",".join(self.all_rates)})
            ET.SubElement(plate, "log", {
                "idref":"featureClockRate:%s:$(rate)" % self.name})
            # Log the scale, but not the shape, as it is always 1 / scale
            # We prefer the scale because it is positively correlated with extent of variation
            ET.SubElement(logger,"log",{"idref":"featureClockRateGammaShape:%s" % self.name})

    def add_frequency_logs(self, logger):
        for f in self.features:
            fname = "%s:%s" % (self.name, f)
            ET.SubElement(logger,"log",{"idref":"feature_freqs_param.s:%s" % fname})



import xml.etree.ElementTree as ET

from .basemodel import BaseModel


class BinaryModel(BaseModel):

    def __init__(self, model_config, global_config):

        BaseModel.__init__(self, model_config, global_config)
        self.remove_constant_features = model_config.get("remove_constant_features", True)
        # Compute feature properties early to facilitate auto-detection of binarisation
        self.compute_feature_properties()
        # Don't need a separating comma because each datapoint is a string
        # of length 1
        self.data_separator = ""
        self.binarised = model_config.get("binarised", None)
        assert type(self.binarised) in (bool, type(None))
        # Do we need to recode multistate data?
        self.recoded = any([self.valuecounts[f]>2 for f in self.features])
        # Check for inconsistent configuration
        if self.recoded and self.binarised:
            raise ValueError("Data for model '%s' contains features with more than two states, but binarised=True was given.  Have you specified the correct data file or feature list?" % self.name)

    def compute_weights(self):
        if not self.recoded:
            BaseModel.compute_weights(self)
        else:
            self.weights = []
            if self.rate_partition:
                for part in sorted(list(set(self.rate_partition.values()))):
                    weight = 0
                    for f in self.features:
                        if self.rate_partition[f] == part:
                            weight += self.valuecounts[f]
                    self.weights.append(weight)
            else:
                for f in self.features:
                    self.weights.append(self.valuecounts[f])

    def set_ascertained(self):
        """
        Decide whether or not to do ascertainment correction for non-constant
        data, unless the user has provided an explicit setting.
        """
        if self.ascertained == None:
            # For binary models, it is possible to retain constant data
            # So we need to be more careful about automatically setting the value
            # of ascertained.
            if self.constant_feature:
                # There's a constant feature in the data, so we definitely shouldn't
                # do ascertainment correction for non-constant data
                self.ascertained = False
            elif self.constant_feature_removed:
                # BEASTling personally removed a constant feature, so we definitely
                # should do ascertainment correction if timing information is
                # important
                self.ascertained = not self.global_config.arbitrary_tree
            else:
                # We didn't see any constant features, but we also didn't remove
                # any.  So we don't quite know what to do...
                # Most data sources are *probably* non-constant, so do ascertainment
                # if the tree is calibrated, but inform the user.
                # This duplicates above, but this condition is a default guess
                # which we may change in future, whereas the above is a logical
                # necessity, so it makes sense to separate these cases
                self.ascertained = not self.global_config.arbitrary_tree
        elif self.ascertained and not self.remove_constant_features:
            raise ValueError("Incompatible settings for model '%s': ascertained=True and remove_constant_features=False together constitute a model misspecification.")
        # If the data has only two values, we need to decide what kind to treat
        # it as
        if not self.recoded:
            if type(self.binarised) == bool:
                self.recoded = self.binarised
            else:
                # Data is binary but we haven't been told if it's "real binary"
                # data or recoded multistate data.  Assume it's real but if
                # constant features have been retained then alert the user
                # because this could cause problems.
                self.recoded = False
                if not self.ascertained:
                    self.messages.append("""[INFO] Model "%s": Assuming that data source %s contains binary structural data (e.g. absence/presence).  If this is cognate set data which has been pre-binarised, please set "binarised=True" in your config to enable appropriate ascertainment correction for the recoding.  If you don't do this, estimates of branch lengths and clade ages may be biased.""" % (self.name, self.data_filename))

    def pattern_names(self, feature):
        """Content of the columns corresponding to this feature in the alignment.

        This method is used for displaying helpful column names in ancestral
        state reconstruction output. It gives column headers for actual value
        columns as well as for dummy columns used in ascertainment correction,
        if such columns exist.

        """
        return (["{:}_dummy{:d}".format(feature, i) for i in range(self.extracolumns[feature])] +
                ["{:}_{:}".format(feature, i) for i in self.unique_values[feature]])

    def format_datapoint(self, feature, point):
        if not self.recoded:
            # This is "true binary" data, and doesn't need to be
            # treated any differently to usual.
            return BaseModel.format_datapoint(self, feature, point)
        else:
            # This is multistate data recoded into binary data.
            if self.ascertained:
                extra_columns = 2
            else:
                # If we are not ascertaining on non-constant data, we still
                # need to add one "all zeros" column to account for the recoding
                extra_columns = 1
            self.extracolumns[feature] = extra_columns
            if point == "?":
                valuestring = "".join(["?" for i in range(0,self.valuecounts[feature]+extra_columns)])
            else:
                # Start with all zeros
                valuestring = ["0" for i in range(0,self.valuecounts[feature] +  extra_columns)]
                # If we're doing full ascertainment we need to set the 2nd extra column to 1
                if self.ascertained:
                    valuestring[1] = "1"
                # Set the appropriate data column to 1
                valuestring[extra_columns + self.unique_values[feature].index(point)] = "1"
                valuestring = "".join(valuestring)
            return valuestring

    def add_feature_data(self, distribution, index, feature, fname):
        data = BaseModel.add_feature_data(self, distribution, index, feature, fname)
        if self.recoded:
            data.set("ascertained", "true")
            data.set("excludefrom", "0")
            if self.ascertained:
                data.set("excludeto", "2")
            else:
                data.set("excludeto", "1")



import sys
import math
import collections

import xml.etree.ElementTree as ET


registered_distributions = (
    ("Beta", "beast.math.distributions.Beta"),
    ("Exponential", "beast.math.distributions.Exponential"),
    ("InverseGamma", "beast.math.distributions.InverseGamma"),
    ("LogNormal", "beast.math.distributions.LogNormalDistributionModel"),
    ("Gamma", "beast.math.distributions.Gamma"),
    ("Uniform", "beast.math.distributions.Uniform"),
    ("LaplaceDistribution", "beast.math.distributions.LaplaceDistribution"),
    ("OneOnX", "beast.math.distributions.OneOnX"),
    ("Normal", "beast.math.distributions.Normal"),
)


DISTRIBUTIONS = {"normal": ("Normal", ("mean", "sigma")),
                 "lognormal": ("LogNormal", ("M", "S")),
                 "uniform": ("Uniform", ("lower", "upper"))}


def add_prior_density_description(compound_distribution, distribution):
    """Create a distribution of the specified type inside an ET element.

    Create an ET sub-element describing a Beast real-parameter
    distribution inside the ET element `compound_distribution`
    reflecting the properties of `distribution`.

    Parameters
    ----------
    compound_distribution: ET.Element
        The xml tag to which the distribution should be added.
    distribution: configuration.Calibration-like
        A description of the distribution.
        Must have a the offset, dist and param attributes.

    Returns
    -------
    None

    Side Effects
    ------------
    Creates a sub-element in `compound_distribution`.
    May register distributions with global map list.

    """
    dist_type, ps = DISTRIBUTIONS[distribution.dist]
    attribs = {"id": "DistributionFor{:}".format(
        compound_distribution.attrib["id"]), "name": "distr", "offset": "0.0"}
    if distribution.offset:
        attribs["offset"] = str(distribution.offset)
    for parameter, value in zip(ps, distribution.param):
        attribs[parameter] = str(value)
    ET.SubElement(compound_distribution, dist_type, attribs)


def parse_prior_string(cs, prior_name="?", is_point=False):
    """Parse a prior-describing string.

    The basic format of such a string is [offset + ]distribution
    Offset is a number, distribution can describe a probability
    density function of normal, lognormal (including rlognormal, a
    reparametrization where the mean is given in real space, not
    in log space) or uniform type in one of the following
    ways. Pseudo-densities with infinite integral are permitted.
    
    Parameters separated using `,` are directly the parameters of the
    distribution. A range separated by a `-` gives the 95% interval of
    that distribution. (This behaviour may change in the future.)

    >>> parse = parse_prior_string
    >>> # Parameters of a normal distribution
    >>> parse("0, 1")
    (0.0, 'normal', (0.0, 1.0))
    >>> # Parameters of some other distribution
    >>> parse(" rlognormal(1, 1)")
    (0.0, 'lognormal', (0.0, 1.0))
    >>> # A distribution shape and its 95%-interval
    >>> parse("normal (1-5)")
    (0.0, 'normal', (3.0, 1.0204081632653061))
    >>> parse("1 - 5")
    (0.0, 'normal', (3.0, 1.0204081632653061))
    >>> parse(">1200")
    (0.0, 'uniform', (1200.0, 9223372036854775807))
    >>> parse("< 1200")
    (0.0, 'uniform', (0.0, 1200.0))

    All of these strings can also be used for point distributions

    >>> parse("normal (1-5)", is_point=True)
    (0.0, 'normal', (3.0, 1.0204081632653061))

    but in addition, point distributions support fixed values.

    >>> parse("300", is_point=True)
    (0.0, 'point', (300.0, 300.0))

    In some cases, in particular for lognormal distributions, it may
    be useful to specify an offset. This is possibly with the syntax

    >>> parse("4 + lognormal(1, 1)")
    (4.0, 'lognormal', (1.0, 1.0))

    The offset must appear *before* the distribution, the other order
    is not permitted.

    >>> parse("lognormal(1, 1) + 4") # doctest: +IGNORE_EXCEPTION_DETAIL
    Traceback (most recent call last):
      File "<stdin>", line 1, in <module>
        parse("lognormal(1, 1) + 4")
      File "beastling/distributions.py", line 148, in parse_prior_string
        offset = float(os.strip())
    ValueError: could not convert string to float: 'lognormal(1, 1)'

    Note
    ====

    For uniform distributions, "uniform(0-1) does not give the 95%
    interval, but the lower and upper bounds.

    While the ">1" and "<1" notations generate uniform
    distributions, the bare "0-1" notation generates a normal
    distribution.

    Parameters
    ==========
    cs: str
        A string describing a single-value probability distribution.
    name: str
        The name of the prior distribution, used only for error
        reporting purposes.
    is_point: bool
        defines whether the distribution is permitted to be a constant
        (point) distribution.

    Returns
    =======
    offset: int
    type: str
        A known distribution type
    parameters: tuple of floats
        The parameters for that distribution

    """
    orig_cs = cs[:]
    # Find offset
    if cs.count("+") == 1:
        os, dist = cs.split("+")
        offset = float(os.strip())
        cs = dist.strip()
    else:
        offset = 0.0

    # Parse distribution
    if cs.count("(") == 1 and cs.count(")") == 1:
        dist_type, cs = cs.split("(", 1)
        dist_type = dist_type.strip().lower()
        if dist_type not in ("uniform", "normal", "lognormal", "rlognormal"):
            raise ValueError(
                "Prior specification '{:}' for {:}"
                " uses an unknown distribution {:}!".format(
                    orig_cs, prior_name, dist_type))
        cs = cs[0:-1]
    else:
        # Default to normal
        dist_type = "normal"

    # Parse / infer params
    if cs.count(",") == 1 and not any([x in cs for x in ("<", ">")]):
        # We've got explicit params
        p1, p2 = map(float, cs.split(","))
    elif cs.count("-") == 1 and not any([x in cs for x in (",", "<", ">")]):
        # We've got a 95% HPD range
        lower, upper = map(float, cs.split("-"))
        if upper <= lower:
            raise ValueError(
                "Prior specification '{:}' for {:}"
                " has an upper bound {:}"
                " which is not higher than its lower bound {:}!".format(
                    orig_cs, prior_name, upper, lower))
        mid = (lower + upper) / 2.0
        if dist_type == "normal":
            p1 = (upper + lower) / 2.0
            p2 = (upper - mid) / 1.96
        elif dist_type == "lognormal":
            p1 = math.log(mid)
            p2a = (p1 - math.log(lower)) / 1.96
            p2b = (math.log(upper) - p1) / 1.96
            p2 = (p2a + p2b) / 2.0
        elif dist_type == "uniform":
            p1 = lower
            p2 = upper
    elif (cs.count("<") == 1 or cs.count(">") == 1) and not any(
            [x in cs for x in (",", "-")]):
        # We've got a single bound
        dist_type = "uniform"
        sign, bound = cs[0], cs[1:]
        if sign == "<":
            p1 = 0.0
            p2 = float(bound.strip())
        elif sign == ">":
            p1 = float(bound.strip())
            p2 = sys.maxsize
        else:
            raise ValueError(
                "Prior specification '{:}' for {:}"
                " cannot be parsed!".format(
                    orig_cs, prior_name))
    elif is_point:
        # Last chance: It's a single language pinned to a
        # single date, so make sure to pin it to that date
        # late and nothing else is left to do with this
        # prior specification.
        try:
            dist_type = "point"
            p1 = float(cs)
            p2 = p1
        except ValueError:
            raise ValueError(
                "Prior specification '{:}' for {:}"
                " cannot be parsed!".format(
                    orig_cs, prior_name))
    else:
        raise ValueError(
            "Prior specification '{:}' for {:}"
            " cannot be parsed!".format(
                orig_cs, prior_name))

    # If this is a lognormal prior specification with the mean in
    # realspace, adjust
    if dist_type == "rlognormal":
        p1 = math.log(p1)
        dist_type = "lognormal"

    # All done!
    return offset, dist_type, (p1, p2)


class Distribution(collections.namedtuple(
        "Distribution", ["offset", "dist", "param"])):
    @classmethod
    def from_string(cls, string, context=None, is_point=False, **kwargs):
        """Create a Distribution object from a prior description string.

        """
        offset, dist, param = parse_prior_string(
            cs=string, prior_name=context, is_point=is_point)
        return cls(offset=offset, dist=dist, param=param, **kwargs)

    def generate_xml_element(self, parent):
        add_prior_density_description(compound_distribution=parent,
                                      distribution=self)

    def mean(self):
        if self.dist in ("normal", "point"):
            return self.offset + self.param[0]
        elif self.dist == "lognormal":
            return self.offset + math.exp(self.param[0])
        elif self.dist == "uniform":
            return self.offset + sum(self.param) / 2.0
        else:
            raise NotImplementedError



import io
import os
import xml.etree.ElementTree as ET

class BaseClock(object):

    def __init__(self, clock_config, global_config):

        self.config = global_config
        # By default, whether or not to estimate the rate is left up to
        # BEASTling.
        default_estimate_rate = None
        # But if the user specifies a rate, we assume they do not want it
        # estimated...
        if any([k in clock_config for k in ("mean", "rate")]):
            k = "mean" if "mean" in clock_config else "rate"
            self.initial_mean = clock_config[k]
            default_estimate_rate = False
        else:
            self.initial_mean = 1.0
        # ...but they can override this by explicitly saying so.
        self.estimate_rate = clock_config.get("estimate_rate",default_estimate_rate)
        self.calibrations = global_config.calibrations
        self.name = clock_config["name"] 
        self.mean_rate_id = "clockRate.c:%s" % self.name
        self.mean_rate_idref = "@%s" % self.mean_rate_id
        self.is_used = False

    def add_state(self, state):
        # Add mean clock rate
        parameter = ET.SubElement(state, "parameter", {
            "id": self.mean_rate_id,
            "upper": "1000.0",
            "name": "stateNode"
            })
        parameter.text=str(self.initial_mean)

    def add_prior(self, prior):
        # Uniform prior on mean clock rate
        sub_prior = ET.SubElement(prior, "prior", {"id":"clockPrior:%s" % self.name, "name":"distribution","x":"@clockRate.c:%s" % self.name})
        uniform = ET.SubElement(sub_prior, "Uniform", {"id":"UniformClockPrior:%s" % self.name, "name":"distr", "upper":"Infinity"})

    def add_branchrate_model(self, beast): # pragma: no cover
        pass

    def add_pruned_branchrate_model(self, distribution, name, tree_id): # pragma: no cover
        raise Exception("This clock is not compatible with PrunedTrees!")

    def add_operators(self, run):
        if self.estimate_rate:
            # Scale mean clock rate
            ET.SubElement(run, "operator", {"id":"clockScaler.c:%s" % self.name, "spec":"ScaleOperator","parameter":"@clockRate.c:%s" % self.name, "scaleFactor":"0.5","weight":"3.0"})

    def add_param_logs(self, logger):
        # Log mean clock rate
        ET.SubElement(logger,"log",{"idref":"clockRate.c:%s" % self.name})

